/******************************************************************
 * Copyright (C) 2006 by SG Software.
 *
 * SG MPFC. Remote interface for MPFC.
 * $Id$ 
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, write to the Free 
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#define __USE_GNU
#include <string.h>
#include <ctype.h>
#define __MPFC_OUTER__
#include "types.h"
#include "command.h"
#include "genp.h"
#include "outp.h"
#include "pmng.h"

#define MAX(a, b) ((a) < (b) ? (b) : (a))

/* Plugins manager */
static pmng_t *remote_pmng = NULL;

/* Plugin description */
static char *remote_desc = "Remote interface for MPFC";

/* Plugin author */
static char *remote_author = "Sergey E. Galanov <sgsoftware@mail.ru>";

/* Logger */
static logger_t *remote_log = NULL;

/* Whether plugin is started */
static bool_t remote_started = FALSE;

/* Thread id */
static pthread_t remote_tid = 0;
static int remote_end_thread_pipe[2] = {-1, -1};

/* MPFC configuration */
static cfg_node_t *remote_cfg = NULL;

/* Server socket */
static int remote_sock;

/* List of clients */
typedef struct tag_remote_client_t
{
	int m_socket;
	char *m_msg_chunk;
	struct tag_remote_client_t *m_next;
} remote_client_t;
static remote_client_t *remote_clients = NULL;

/* Message type */
typedef struct 
{
	/* Type of message */
	enum
	{
		REMOTE_MSG_COMMAND,
		REMOTE_MSG_GET
	} m_type;

	/* Complete message */
	char *m_str;

	/* Command in case of COMMAND message or 
	 * object to get in case of GET message */
	char *m_obj;

	/* Parameters */
	cmd_params_list_t *m_params;
} remote_msg_t;

/* Free message */
static void remote_msg_free( remote_msg_t *msg )
{
	if (msg == NULL)
		return;

	if (msg->m_str != NULL)
		free(msg->m_str);
	if (msg->m_obj != NULL)
		free(msg->m_obj);
	if (msg->m_params != NULL)
		cmd_free_params(msg->m_params);
	free(msg);
} /* End of 'remote_msg_free' function */

/* Parse one message parameter */
static bool_t remote_msg_parse_param( cmd_param_t *param, char **str )
{
	char *s = *str;
	
	/* Skip whitespace */
	while ((*s) == ' ')
		s ++;

	/* String parameter */
	if ((*s) == '\"')
	{
		bool_t escaped = FALSE;
		char *temp, *p, *pp;

		logger_debug(remote_log, "remote: parsing param in %s", s);

		/* Find closing double quote */
		for ( temp = ++s;; )
		{
			int c;
			char *q = strchr(temp, '\"');
			if (q == NULL)
				return FALSE;
			temp = q + 1;

			/* If number of backslashes before this double quote is
			 * even then it is really a closing quote */
			for ( c = 0, q --; (*q) == '\\'; q --, c ++ );
			logger_debug(remote_log, "remote: c is %d", c);
			if ((c % 2) == 1)
				continue;
			else
				break;
		}

		/* Allocate memory */
		p = (char *)malloc(temp - s);
		if (p == NULL)
		{
			logger_debug(remote_log, "remote: no enough memory for message");
			return FALSE;
		}

		/* Read */
		for ( pp = p;; s ++ )
		{
			if ((*s) == '\\' && !escaped)
			{
				escaped = TRUE;
				continue;
			}
			else if ((*s) == '\"' && !escaped)
			{
				(*pp) = 0;
				s ++;
				break;
			}

			(*pp ++) = (*s);
		}
		logger_debug(remote_log, "remote: parameter is string %s; s is %s", p, s);
		param->m_type = CMD_PARAM_STRING;
		param->m_value.m_string = p;
		(*str) = s;
		return TRUE;
	}
	/* Integer */
	else if (isdigit(*s))
	{
		int num = 0;
		for ( ; isdigit(*s); s ++ )
		{
			num *= 10;
			num += ((*s) - '0');
		}
		logger_debug(remote_log, "remote: parameter is int %d; s is %s", num, s);
		param->m_type = CMD_PARAM_INT;
		param->m_value.m_int = num;
		(*str) = s;
		return TRUE;
	}

	/* Unknown */
	return FALSE;
} /* End of 'remote_msg_parse_param' function */

/* Parse message parameters */
static bool_t remote_msg_parse_params( remote_msg_t *msg, char *s )
{
	cmd_params_list_t *pl;

	logger_debug(remote_log, "remote: parsing message parameters in %s", s);

	/* Allocate memory */
	pl = (cmd_params_list_t *)malloc(sizeof(*pl));
	if (pl == NULL)
	{
		logger_error(remote_log, 0, "remote: no enough memory for message");
		return FALSE;
	}
	pl->m_params = NULL;
	pl->m_num_params = 0;
	pl->m_iterator = 0;
	msg->m_params = pl;

	for ( ;; )
	{
		cmd_param_t p;
		cmd_param_t *ppl;

		if (!remote_msg_parse_param(&p, &s))
			break;

		ppl = (cmd_param_t *)realloc(pl->m_params,
				(sizeof(p) * (pl->m_num_params + 1)));
		if (ppl == NULL)
		{
			logger_error(remote_log, 0, "remote: no enough memory for message");
			return FALSE;
		}
		pl->m_params = ppl;
		pl->m_params[pl->m_num_params ++] = p;
	}
	return TRUE;
} /* End of 'remote_msg_parse_params' function */

/* Parse message */
static remote_msg_t *remote_msg_parse( char *s )
{
	remote_msg_t *msg = NULL;
	char *p;

	logger_debug(remote_log, "remote: parsing message %s", s);

	/* Allocate memory */
	msg = (remote_msg_t *)malloc(sizeof(*msg));
	if (msg == NULL)
	{
		logger_error(remote_log, 0, "remote: no enough memory for message");
		return NULL;
	}
	memset(msg, 0, sizeof(*msg));
	msg->m_str = s;

	/* Get message type */
	if (!strncmp(s, "command ", 8))
	{
		msg->m_type = REMOTE_MSG_COMMAND;
		s += 8;
	}
	else if (!strncmp(s, "get ", 4))
	{
		msg->m_type = REMOTE_MSG_GET;
		s += 4;
	}
	else
	{
		logger_error(remote_log, 0, "remote: unknown message type");
		remote_msg_free(msg);
		return NULL;
	}
	logger_debug(remote_log, "remote: message type is %d, s is %s",
			msg->m_type, s);

	/* Get object */
	for ( p = s; *p && (*p != '\n') && (*p != '\r') && (*p != ' '); p ++ );
	msg->m_obj = strndup(s, p - s);
	s = p + 1;
	if (msg->m_obj == NULL)
	{
		logger_error(remote_log, 0, "remote: no enough memory for message");
		remote_msg_free(msg);
		return NULL;
	}
	logger_debug(remote_log, "remote: message object is %s, s is %s",
			msg->m_obj, s);

	/* Parse parameters */
	if (!remote_msg_parse_params(msg, s))
	{
		logger_error(remote_log, 0, "remote: unable to parse message parameters");
		remote_msg_free(msg);
		return NULL;
	}

	return msg;
} /* End of 'remote_msg_parse' function */

/* Add a client */
static bool_t remote_clients_add( int s )
{
	remote_client_t *client;

	/* Initialize client */
	logger_debug(remote_log, "remote: adding client %d", s);
	client = (remote_client_t *)malloc(sizeof(*client));
	if (client == NULL)
	{
		logger_error(remote_log, 0, "remote: no enough memory for client");
		return FALSE;
	}
	client->m_socket = s;
	client->m_msg_chunk = NULL;
	client->m_next = remote_clients;
	remote_clients = client;
	return TRUE;
} /* End of 'remote_clients_add' function */

/* Delete a client from list */
static void remote_clients_delete( remote_client_t *client )
{
	remote_client_t *t;

	logger_debug(remote_log, "remote: deleting client %d", client->m_socket);

	/* Close connection */
	close(client->m_socket);
	if (client->m_msg_chunk != NULL)
		free(client->m_msg_chunk);

	/* Client is the first element of the list */
	if (client == remote_clients)
	{
		free(client);
		remote_clients = NULL;
		return;
	}
	
	/* Search for this client in the list */
	for ( t = remote_clients; t != NULL && t->m_next != client; t = t->m_next );
	if (t == NULL)
		return;

	/* Client found */
	t->m_next = client->m_next;
	free(client);
} /* End of 'remote_clients_delete' function */

/* Free clients */
static void remote_clients_free( void )
{
	remote_client_t *client;

	for ( client = remote_clients; client != NULL; )
	{
		remote_client_t *next = client->m_next;
		close(client->m_socket);
		if (client->m_msg_chunk != NULL)
			free(client->m_msg_chunk);
		free(client);
		client = next;
	}
	remote_clients = NULL;
} /* End of 'remote_clients_free' function */

/* Read a message from client */
static remote_msg_t *remote_clients_read_msg( remote_client_t *client )
{
	char buf[1024];
	int len = 0;
	char *msg = NULL;

	logger_debug(remote_log, "remote: in read message with chunk %s",
			client->m_msg_chunk);
	msg = client->m_msg_chunk;
	if (msg == NULL)
		msg = strdup("");
	len = strlen(msg);

	/* Receive message that is currently available */
	for ( ;; )
	{
		int l;

		/* Receive part of message */
		l = recv(client->m_socket, buf, sizeof(buf), MSG_DONTWAIT);
		logger_debug(remote_log, "remote: received %s(%d)", buf, l);
		if (l < 0)
		{
			free(msg);
			return NULL;
		}
		logger_debug(remote_log, "remote: reallocating message of length %d "
				"(previous was %p of len %d", len + l + 1, msg, len + 1);
		msg = (char *)realloc(msg, len + l + 1);
		logger_debug(remote_log, "remote: after reallocation msg is %p", msg);
		strncat(msg, buf, l);
		len += l;
		logger_debug(remote_log, "remote: after reallocation len is %d", len);
		if (l < sizeof(buf))
			break;
	}
	logger_debug(remote_log, "remote: after cycle message is %s", msg);

	/* No message received: means that client has terminated */
	if (len == 0)
	{
		logger_debug(remote_log, "remote: client terminated");
		remote_clients_delete(client);
		return NULL;
	}

	/* Complete message must end with an endline */
	if (msg[len - 1] != '\n')
	{
		logger_debug(remote_log, "remote: message is incomplete");
		client->m_msg_chunk = msg;
		return NULL;
	}

	/* Message is complete */
	client->m_msg_chunk = NULL;

	/* Parse message */
	return remote_msg_parse(msg);
} /* End of 'remote_clients_read_msg' function */

/* Send response to client */
static void remote_clients_send_response( remote_client_t *client, char *response, ... )
{
	char s[1024], *p;
	va_list ap;
	int len;

	va_start(ap, response);
	vsnprintf(s, sizeof(s), response, ap);
	va_end(ap);
	logger_debug(remote_log, "sending response: %s", s);
	for ( len = strlen(s), p = s; len > 0; )
	{
		int ret = send(client->m_socket, p, len, 0);
		logger_debug(remote_log, "send returned %d", ret);
		if (ret < 0)
			return;
		p += ret;
		len -= ret;
	}
} /* End of 'remote_clients_send_response' function */

/* Handle message */
static void remote_handle_msg( remote_client_t *client, remote_msg_t *msg )
{
	logger_debug(remote_log, "remote: handling message %s with object %s", 
			msg->m_str, msg->m_obj);

	/* A command */
	if (msg->m_type == REMOTE_MSG_COMMAND)
	{
		pmng_player_command_obj(remote_pmng, msg->m_obj, msg->m_params);
		msg->m_params = NULL;
		remote_clients_send_response(client, "ok\n");
	}
	/* Get */
	else if (msg->m_type == REMOTE_MSG_GET)
	{
		logger_debug(remote_log, "get message object is %s", msg->m_obj);
		if (!strcmp(msg->m_obj, "playlist"))
		{
			plist_t *pl = remote_pmng->m_playlist;
			remote_clients_send_response(client, 
					"Status: ok\nSize: %d\nCurrent: %d\n\n", 
					pl->m_len, pl->m_cur_song);
		}
		else if (!strcmp(msg->m_obj, "song"))
		{
			int index = cmd_next_int_param(msg->m_params);
			plist_t *pl = remote_pmng->m_playlist;
			song_t *s;

			logger_debug(remote_log, "getting song with index %d", index);
			if (index < 0 || index >= pl->m_len)
			{
				remote_clients_send_response(client,
						"Status: failed\nError: invalid index\n\n");
				return;
			}

			s = pl->m_list[index];
			remote_clients_send_response(client,
					"Status: ok\nTitle: %s\nName: %s\nLength: %d\n\n",
					STR_TO_CPTR(s->m_title), s->m_file_name, s->m_len);
		}
		else if (!strcmp(msg->m_obj, "player"))
		{
			player_context_t *pc = remote_pmng->m_player_context;
			remote_clients_send_response(client,
					"Status: ok\nState: %s\nTime: %d\n"
					"Bitrate: %d\nFreq: %d\nStereo: %d\n"
					"Volume: %d\nBalance: %d\n\n",
					(pc->m_status == PLAYER_STATUS_PLAYING ? "playing" : 
					 (pc->m_status == PLAYER_STATUS_PAUSED ? "paused" : 
					  "stopped")),
					pc->m_cur_time, pc->m_bitrate, pc->m_freq, pc->m_stereo,
					pc->m_volume, pc->m_balance);
		}
		else if (!strcmp(msg->m_obj, "volume-full"))
		{
			int left, right;
			outp_get_volume(remote_pmng->m_cur_out, &left, &right);
			remote_clients_send_response(client,
					"Status: ok\nLeft: %d\nRight: %d\n\n", left, right);
		}
		else
		{
			remote_clients_send_response(client, 
					"Status: failed\nError: unknown object\n\n");
		}
	}
	else
		remote_clients_send_response(client, "failed\n");
} /* End of 'remote_handle_msg' function */

/* Main thread */
static void *remote_thread( void *arg )
{
	int end_thread = remote_end_thread_pipe[0];
	remote_client_t *client;

	logger_debug(remote_log, "remote: in remote_thread with socket %d", 
			remote_sock);
	for ( ;; )
	{
		fd_set rdfs;
		int ret, m;

		/* Set up file sets */
		FD_ZERO(&rdfs);
		FD_SET(remote_sock, &rdfs);
		FD_SET(end_thread, &rdfs);
		m = MAX(remote_sock, end_thread);
		for ( client = remote_clients; client != NULL; client = client->m_next )
		{
			int s = client->m_socket;
			logger_debug(remote_log, "remote: adding client %d to the set", s);
			FD_SET(s, &rdfs);
			if (s > m)
				m = s;
		}

		/* Check for new connections or messages from clients */
		ret = select(m + 1, &rdfs, NULL, NULL, NULL);
		logger_debug(remote_log, "remote: select returned %d", ret);
		if (ret < 0)
			continue;

		/* End thread */
		if (FD_ISSET(end_thread, &rdfs))
			break;

		/* New connection */
		if (FD_ISSET(remote_sock, &rdfs))
		{
			int s;

			/* Accept connection */
			logger_debug(remote_log, "remote: new connection");
			s = accept(remote_sock, NULL, NULL);
			logger_debug(remote_log, "remote: accepted socket %d", s);
			if (s < 0)
			{
				logger_error(remote_log, 0, "remote: unable to accept connection");
			}
			/* Save client */
			else if (!remote_clients_add(s))
			{
				logger_error(remote_log, 0, "remote: unable to add client");
			}
		}

		/* Read messages from clients */
		for ( client = remote_clients; client != NULL; client = client->m_next )
		{
			remote_msg_t *msg;

			if (!FD_ISSET(client->m_socket, &rdfs))
				continue;

			/* Read and handle message */
			logger_debug(remote_log, "remote: got something from %d", 
					client->m_socket);
			msg = remote_clients_read_msg(client);
			if (msg != NULL)
			{
				logger_debug(remote_log, "remote: got a complete message");
				remote_handle_msg(client, msg);
				remote_msg_free(msg);
			}
			else
			{
				remote_clients_send_response(client, "failed\n");
			}
		}
	}

	/* Close connections */
	remote_clients_free();
	return NULL;
} /* End of 'remote_thread' function */

/* Start plugin */
bool_t remote_start( void )
{
	int ret, port, yes = 1;
	struct sockaddr_in name;

	logger_debug(remote_log, "remote: in remote_start");

	/* Create server socket */
	remote_sock = socket(PF_INET, SOCK_STREAM, 0);
	if (remote_sock < 0)
	{
		logger_error(remote_log, 0, "remote: socket creation failed");
		return FALSE;
	}
	setsockopt(remote_sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
	if (fcntl(remote_sock, F_SETFL, O_NONBLOCK) == -1)
	{
		logger_error(remote_log, 0, "remote: unable to set non-block flag on socket");
		close(remote_sock);
		remote_sock = -1;
		return FALSE;
	}
	name.sin_family = AF_INET;
	port = cfg_get_var_int(remote_cfg, "port");
	if (port == 0)
		port = 40000;
	name.sin_port = htons(port);
	name.sin_addr.s_addr = INADDR_ANY;
	if (bind(remote_sock, (struct sockaddr *)&name, sizeof(name)) < 0)
	{
		logger_error(remote_log, 0, "remote: bind failed");
		close(remote_sock);
		remote_sock = -1;
		return FALSE;
	}
	if (listen(remote_sock, 10) < 0)
	{
		logger_error(remote_log, 0, "remote: listen failed");
		close(remote_sock);
		remote_sock = -1;
		return FALSE;
	}

	/* Create pipe for ending thread */
	if (pipe(remote_end_thread_pipe) < 0)
	{
		logger_error(remote_log, 0, "remote: pipe failed");
		close(remote_sock);
		remote_sock = -1;
		return FALSE;
	}

	/* Start server thread */
	ret = pthread_create(&remote_tid, NULL, remote_thread, NULL);
	if (ret < 0)
	{
		logger_error(remote_log, 0, "remote: pthread_create failed");
		return FALSE;
	}
	remote_started = TRUE;
	return TRUE;
} /* End of 'remote_start' function */

/* End plugin */
void remote_end( void )
{
	remote_started = FALSE;

	if (remote_tid != 0)
	{
		write(remote_end_thread_pipe[1], "end", 3);
		pthread_join(remote_tid, NULL);
		remote_tid = 0;
		close(remote_end_thread_pipe[0]);
		close(remote_end_thread_pipe[1]);
	}

	if (remote_sock >= 0)
	{
		close(remote_sock);
		remote_sock = -1;
	}
} /* End of 'remote_end' function */

/* Check whether plugin is started */
bool_t remote_is_started( void )
{
	return remote_started;
} /* End of 'remote_is_started' function */

/* Exchange data with main program */
void plugin_exchange_data( plugin_data_t *pd )
{
	pd->m_desc = remote_desc;
	pd->m_author = remote_author;
	GENP_DATA(pd)->m_start = remote_start;
	GENP_DATA(pd)->m_is_started = remote_is_started;
	GENP_DATA(pd)->m_end = remote_end;
	remote_pmng = pd->m_pmng;
	remote_log = pd->m_logger;
	remote_cfg = pd->m_cfg;
} /* End of 'plugin_exchange_data' function */

/* End of 'remote.c' file */

